import pytest
from speedyfit.photometry_query import get_parallax


def test_get_parallax():

    plx, e_plx = get_parallax('BD+34 1543', radius=3)

    assert plx == pytest.approx(5.172398, 0.000001)
    assert e_plx == pytest.approx(0.0397, 0.0001)




from astropy.coordinates import Angle


def convert_ra(ra):
    if ' ' in str(ra).strip() or ':' in str(ra).strip():
        ra = str(ra).strip()
        ra = ra.replace(' ', '').replace(':', '')
    else:
        a = Angle(ra, unit='degree').hms
        ra = '{:02.0f}{:02.0f}{:05.2f}'.format(*a)
    return ra


def convert_dec(dec):
    if ' ' in str(dec).strip() or ':' in str(dec).strip():
        dec = str(dec).strip()
        dec = dec.replace(' ', '').replace(':', '')
        if not '-' in dec and not '+' in dec:
            dec = '+' + dec
    else:
        a = Angle(dec, unit='degree').dms
        dec = '{:+03.0f}{:02.0f}{:05.2f}'.format(a[0], abs(a[1]), abs(a[2]))
    return dec


def get_name_from_coordinates(ra, dec):

    if hasattr(ra, '__iter__'):
        ra_ = [convert_ra(x) for x in ra]
        dec_ = [convert_dec(x) for x in dec]
        name = [f'J{r}{d}' for r, d in zip(ra_, dec_)]
    else:
        ra_, dec_ = convert_ra(ra), convert_dec(dec)
        name = f'J{ra_}{dec_}'

    return name


import os
import sys
import time
import shutil

import numpy as np

from astropy.io import fits
from astropy.table import Table

from multiprocessing import cpu_count, Manager, Process

from speedyfit import model, reddening, filters


def get_responses(responses=None, wave=(0, np.inf)):
    """
    Get a list of response functions and their information.

    You can specify bandpass systems (e.g. GENEVA) and then all Geveva filters
    will be collected. You can also specific specific bandpasses (e.g. GENEVA.V),
    in which case only that one will be returned. The default C{None} returns
    all systems except some Hubble ones.

    You can also set responses to C{None} and give a wavelength range for filter
    selection.

    Example input for C{responses} are:

    >>> responses = ['GENEVA.V','2MASS.J']
    >>> responses = ['GENEVA','2MASS.J']
    >>> responses = ['BOXCAR','STROMGREN']
    >>> responses = None

    @param responses: a list of filter systems of passbands
    @type responses: list of str
    @param add_spectrophotometry: add spectrophotometric filters to the filter list
    @type add_spectrophotometry: bool
    @param wave: wavelength range
    @type wave: tuple (float,float)
    """

    # -- if no responses are given, select using wavelength range
    if responses is None:
        responses = filters.list_response(wave_range=(wave[0], wave[-1]))
    else:
        responses_ = []
        for resp in responses:
            print('... subselection: {}'.format(resp))
            responses_ += filters.list_response(resp)
        responses = responses_
    # -- get information on the responses
    responses = [resp for resp in responses if not (
            ('ACS' in resp) or ('WFPC' in resp) or ('STIS' in resp) or ('ISOCAM' in resp) or ('NICMOS' in resp))]

    print('Selected response curves: {}'.format(', '.join(responses)))

    return responses


def get_threads(threads, max=np.inf):
    """
    Reads the threadcount, and returns an integer
    accepts: <integer>, 'max', 'half', 'safe'
       max: all cpus are used
       half: half of the cpus are used
       safe: all but one of the cpus are used
    """
    if threads == 'max':
        threads = cpu_count()
    elif threads == 'half':
        threads = cpu_count() / 2
    elif threads == 'safe':
        threads = cpu_count() - 1
    threads = int(threads)

    if threads > max:
        threads = max

    return threads


# def calc_integrated_grid(threads=1, ebvs=None, law='fitzpatrick2004', Rv=3.1, responses=None, update=False, grid=None):
#
#     # get the teff and logg combinations of the grid
#     teffs, loggs = model.get_grid_dimensions(grid=grid)
#
#     def process_ebvs(wave, flux, ebvs):
#
#         arr = []
#
#         for ebv in ebvs:
#             # redden the flux
#             flux_ = reddening.redden(flux, wave=wave, ebv=ebv, rtype='flux', law=law, Rv=Rv)
#
#             # calculate synthetic fluxes
#             synflux = model.synthetic_flux(wave, flux_, responses)
#
#             arr.append([np.concatenate(([ebv], synflux))])
#
#         return arr
#
#     for teff, logg in zip(teffs, loggs):
#
#         # get the spectrum
#         wave, flux = model.get_table(grid=grid, teff=teff, logg=logg)
#
#         results = process_ebvs(wave, flux, ebvs)

        

def calc_integrated_grid(threads=1, ebvs=np.r_[0:2.01:0.01], law='fitzpatrick2004', Rv=3.1, responses=None, grid=None):
    """
    Integrate an entire SED grid over all passbands and save to a FITS file.

    The output file can be used to fit SEDs more efficiently, since integration
    over the passbands has already been carried out.

    WARNING: this function can take a long time to compute!

    Extra keywords can be used to specify the grid.

    :param threads: number of threads
    :type threads; integer, 'max', 'half' or 'safe'
    :param ebvs: reddening parameters to include
    :type ebvs: numpy array
    :param law: interstellar reddening law to use
    :type law: string (valid law name, see C{reddening.py})
    :param Rv: Rv value for reddening law
    :type Rv: float
    :param responses: respons curves to add (if None, add all)
    :type responses: list of strings
    """

    # -- select number of threads
    threads = get_threads(threads, max=len(ebvs))

    # -- Get the grid, extract the different teff and logg combinations provided and load the first model to check
    #    which photbands can be calculated.
    gridfile = model.get_grid_file(integrated=False, grid=grid)
    ff = fits.open(gridfile)
    data = ff[1].data
    wave, flux = data['wavelength'], data['flux']

    teffs = []
    loggs = []
    hdus = []
    for hdu in ff[1:]:
        teffs.append(float(hdu.header['TEFF']))
        loggs.append(float(hdu.header['LOGG']))
        hdus.append(hdu)

    responses = get_responses(responses=responses, wave=wave)

    # -- definition of one process for multi processing:
    def do_ebv_process(ebvs, arr, responses):
        # Run over all reddening values, calculate the reddened model and integrate it over all photbands.
        for ebv in ebvs:
            # redden the model
            flux_ = reddening.redden(flux, wave=wave, ebv=ebv, rtype='flux', law=law, Rv=Rv)

            # calculate synthetic fluxes
            synflux = filters.synthetic_flux(wave, flux_, responses)

            # append to results
            arr.append([np.concatenate(([ebv], synflux))])

    # -- prepare the array containing the integrated fluxes
    #   (1 row per model, 1 column for each response curve and teff, logg, ebv and total luminosity)
    output = np.zeros((len(teffs) * len(ebvs), 4 + len(responses)))
    start = 0

    # also track possible errors
    exceptions = 0
    exceptions_logs = []

    print('Total number of tables: %i ' % (len(teffs)))
    c0 = time.time()
    for i, (teff, logg, hdu) in enumerate(zip(teffs, loggs, hdus)):
        if i > 0:
            print('%s %s %s %s: ET %d seconds' % (teff, logg, i, len(teffs), (time.time() - c0) / i * (len(teffs) - i)))

        # -- get model SED and absolute luminosity
        wave, flux = hdu.data['wavelength'], hdu.data['flux']
        Labs = model.luminosity(wave, flux)

        # -- threaded calculation over all E(B-V)s
        manager = Manager()
        arr = manager.list([])
        all_processes = []
        for j in range(threads):
            all_processes.append(Process(target=do_ebv_process, args=(ebvs[j::threads], arr, responses)))
            all_processes[-1].start()
        for p in all_processes:
            p.join()

        try:
            # -- collect the results and add them to 'output'
            arr = np.vstack([row for row in arr])
            sa = np.argsort(arr[:, 0]) # sort on ebv
            arr = arr[sa]
            output[start:start + arr.shape[0], :3] = teff, logg, Labs
            output[start:start + arr.shape[0], 3:] = arr
            start += arr.shape[0]
        except:
            print('Exception in calculating Teff=%f, logg=%f' % (teff, logg))
            print('Exception: %s' % (sys.exc_info()[1]))
            exceptions = exceptions + 1
            exceptions_logs.append(sys.exc_info()[1])

    # -- close the fits file with the model
    ff.close()

    # -- create Table object and add header info to the table
    output = Table(data=output, names=['teff', 'logg', 'Labs', 'ebv'] + responses)
    output.meta['gridfile'] = (os.path.basename(gridfile), 'original model file')
    output.meta['GRID'] = (grid, 'name of the model grid')
    output.meta['fluxtype'] = ('Flambda', 'units of the flux')
    output.meta['redlaw'] = (law, 'interstellar reddening law')
    output.meta['rv'] = (Rv, 'interstellar reddening parameter')

    # -- create the name of the output file and safe to disk
    outfile = 'i{0}'.format(os.path.basename(gridfile))
    outfile = os.path.splitext(outfile)
    outfile = outfile[0] + '_law{0}_Rv{1:.2f}'.format(law, Rv) + outfile[1]
    if os.path.isfile(outfile):
        print('Precaution: making original grid backup at {0}.backup'.format(outfile))
        shutil.copy(outfile, outfile + '.backup')
    output.write(outfile, overwrite=True)

    # # -- make FITS columns
    # output = output.T
    # cols = [fits.Column(name='teff', format='E', array=output[0]),
    #         fits.Column(name='logg', format='E', array=output[1]),
    #         fits.Column(name='ebv', format='E', array=output[3]),
    #         fits.Column(name='Labs', format='E', array=output[2])]
    # for i, photband in enumerate(responses):
    #     cols.append(fits.Column(name=photband, format='E', array=output[4 + i]))
    #
    # # -- make FITS extension and write grid/reddening specifications to header
    # table = fits.TableHDU.from_columns(fits.ColDefs(cols))
    # table.header.update(gridfile=os.path.basename(gridfile))
    # table.header.update(GRID=(grid, 'name of the model grid'),
    #                     FLUXTYPE=('Flambda', 'units of the flux'),
    #                     REDLAW=(law, 'interstellar reddening law'),
    #                     RV=(Rv, 'interstellar reddening parameter'))
    # # -- make/update complete FITS file
    # if os.path.isfile(outfile):
    #     os.remove(outfile)
    #     print('Removed existing file: %s' % (outfile))

    # hdulist = fits.HDUList([])
    # hdulist.append(fits.PrimaryHDU(np.array([[0, 0]])))
    # hdulist.append(table)
    # hdulist.writeto(outfile)
    print("Written output to %s" % outfile)

    print('Encountered %s exceptions!' % exceptions)
    for i in exceptions_logs:
        print('ERROR\n', i)

    return outfile


def check_grid(grid):
    """
    Check if the grid with integrated photometry calculated with calc_integrated_grid
    has all header information, and if all models succeeded.
    """
    print('Checking grid: {}'.format(grid))

    hdulist = fits.open(grid, mode='update')
    names = hdulist[1].columns.names
    for i, name in enumerate(names):
        if name.lower() in ['teff', 'logg', 'ebv', 'labs', 'vrad', 'rv', 'z']:
            names[i] = name.lower()
    cols = [fits.Column(name=name, format='E', array=hdulist[1].data.field(name)) for name in names]
    N = len(hdulist[1].data)

    keys = [key.lower() for key in hdulist[1].header.keys()]

    if 'z' not in hdulist[1].header:
        hdulist[1].header['Z'] = 0.0
        print('Adding metallicity (Z={}) to header!'.format(hdulist[1].header['Z']))

    if 'Rv' not in hdulist[1].header:
        hdulist[1].header['Rv'] = 3.1
        print('Adding Rv (Rv={}) to header!'.format(hdulist[1].header['Rv']))

    if 'z' not in names:
        z = hdulist[1].header.get('z', 0.0)
        print('Adding metallicity from header {}'.format(z))
        cols.append(fits.Column(name='z', format='E', array=np.ones(N) * z))
    else:
        print("Metallicity already in there")
    if 'vrad' not in names:
        vrad = 0.
        print('Adding radial velocity {}'.format(vrad))
        cols.append(fits.Column(name='vrad', format='E', array=np.ones(N) * vrad))
    else:
        print("Radial velocity already in there")

    fix_rv = False
    if 'rv' not in names:
        if 'rv' in keys:
            rv = hdulist[1].header['Rv']
            print("Adding interstellar Rv from header {}".format(rv))
        else:
            rv = 3.1
            print("Adding default interstellar Rv {}".format(rv))
        cols.append(fits.Column(name='rv', format='E', array=np.ones(N) * rv))
    elif not hdulist[1].header['Rv'] == hdulist[1].data.field('rv')[0]:
        rv = hdulist[1].header['Rv']
        fix_rv = rv
        print('Correcting interstellar Rv with {}'.format(rv))
    else:
        print("Interstellar Rv already in there")

    table = fits.TableHDU.from_columns(fits.ColDefs(cols))
    if fix_rv:
        table.data.field('rv')[:] = rv
    fake_keys = [key.lower() for key in table.header.keys()]

    # make sure all keywords set above are also in the new table header
    table.header.update(hdulist[1].header)

    # now update header keywords related to the created table
    for key in hdulist[1].header.keys():
        if not key.lower() in fake_keys:
            if len(key) > 8:
                key = 'HIERARCH ' + key
            table.header.update(key=hdulist[1].header[key])
    hdulist[1] = table

    logstring = "Axis:\n"
    for name in hdulist[1].columns.names:
        if name.islower() and not name == 'labs':
            ax = np.unique(hdulist[1].data.field(name))
            logstring += "\t\t{} {} {} {}\n".format(name, len(ax), min(ax), max(ax))
    print(logstring)

    keep = hdulist[1].data.field('teff') > 0
    print('Removing {}/{} false entries'.format(sum(~keep), len(keep)))
    hdulist[1].data = hdulist[1].data[keep]
    hdulist.flush()
    hdulist.close()


if __name__=="__main__":

    responses = ['GALEX', 'IUE', 'STROMGREN', 'JOHNSON', 'GAIA3E', 'GAIA2', 'SKYMAPPER', 'APASS', 'SDSS', '2MASS', 'WISE']
    evbs = np.r_[0:1.02:0.05]

    # Blackbody
    # calc_integrated_grid(threads=6, ebvs=evbs, law='fitzpatrick2004', Rv=3.1,
    #                      responses=responses, grid='blackbody')

    # # TMAP
    # calc_integrated_grid(threads=6, ebvs=evbs, law='fitzpatrick2004', Rv=3.1,
    #                      responses=responses, grid='tmap')

    # # Kurucz
    # calc_integrated_grid(threads=6, ebvs=evbs, law='fitzpatrick2004', Rv=3.1,
    #                      responses=responses, grid='kurucz')

    # # Munari
    # calc_integrated_grid(threads=6, ebvs=evbs, law='fitzpatrick2004', Rv=3.1,
    #                      responses=responses, grid='munari')

    # Koester
    calc_integrated_grid(threads=6, ebvs=evbs, law='fitzpatrick2004', Rv=3.1,
                         responses=responses, grid='koester')


import copy
import os
import argparse

import yaml

import pandas as pd
import numpy as np

from speedyfit import photometry_query
from speedyfit.main import fit_sed, get_observations, write_results, plot_results
from speedyfit.default_setup import default_tmap, default_binary
from speedyfit.utils import get_name_from_coordinates

def extract_constraint_names(data):
    columns = data.columns.values
    constraint_names = []
    for name in columns:
        if 'constraint_' in name:
            cname = name.split('_')[1]
            constraint_names.append(cname)

    return constraint_names


def process_objects(data):

    # check if we are working with names or if we need to make names from the coordinates
    if 'name' in data.columns.values:
        object_list = data['name'].apply(lambda x: x.replace(' ', '_').strip()).values
        object_list = pd.DataFrame({'name': object_list})
    else:
        # convert coordinates into a name if no name was given.
        name = get_name_from_coordinates(data['ra'].values, data['dec'].values)
        object_list = pd.DataFrame({'name': name})

    if 'ra' in data and 'dec' in data:
        object_list['ra'] = data['ra']
        object_list['dec'] = data['dec']

    # now check for constraints
    constraint_cols = extract_constraint_names(data)
    for cname in constraint_cols:
        object_list[cname] = data['constraint_'+cname]

    return object_list


def read_setup(setup):

    if os.path.isfile(setup):
        infile = open(setup, 'r')
        setup = infile.readlines()
        infile.close()
        setup = ''.join(setup)

    elif setup == 'single':
        setup = default_tmap
    elif setup == 'binary':
        setup = default_binary
    else:
        print('Setup not recognized! Using single as default')
        setup = default_tmap

    return setup


def prepare_setup(object_list, basedir, default_setup):

    constraint_names = list(object_list.columns.values)
    constraint_names.remove('name')
    constraint_names.remove('ra')
    constraint_names.remove('dec')

    for i, row in object_list.iterrows():
        objectname = row['name']
        ra, dec = row['ra'], row['dec']
        print(f'Preparing setup files for object: {objectname} ({ra}  {dec})')

        if not os.path.isdir(basedir+'/'+objectname):
            os.mkdir(basedir+'/'+objectname)

        plx, e_plx = photometry_query.get_parallax(objectname)

        out = copy.copy(default_setup)
        if '<photfilename>' in out:
            photfilename_ = basedir + '/' + objectname + '/' + objectname + '.phot'
            out = out.replace('<photfilename>', photfilename_)
        if '<objectname>' in out:
            objectname_ = basedir + '/' + objectname + '/' + objectname
            out = out.replace('<objectname>', objectname_)
        if '<constraints>' in out:
            constraints = f"\n  parallax: [{plx:0.4f}, {e_plx:0.4f}]"
            for cname in constraint_names:
                val = row[cname]
                try:
                    val = val.replace('(', '').replace(')', '')
                except:
                    continue
                if val.strip() == '':
                    continue
                constraints += f"\n  {cname}: [{val}]"
            out = out.replace('<constraints>', constraints)

        filename = basedir + '/' + objectname + '/' + objectname + '_setup.yaml'

        ofile = open(filename, 'w')
        ofile.write(out)
        ofile.close()


def prepare_photometry(object_list, basedir, skip_existing=True):

    for i, row  in object_list.iterrows():
        objectname = row['name']

        if os.path.isfile(basedir+'/'+objectname+'/'+objectname+'.phot') and skip_existing:
            continue

        if not os.path.isdir(basedir+'/'+objectname):
            os.mkdir(basedir+'/'+objectname)

        try:
            filename = basedir + '/' + objectname + '/' + objectname + '.phot'
            photometry = photometry_query.get_photometry(objectname, filename=filename)
        except Exception as e:
            print("Failed to obtain photometry for {}".format(objectname))
            print(e)


def fit_seds(object_list, basedir, start_index=0, stop_index=None):

    all_results = {}

    stop_index = len(object_list) if stop_index is None else stop_index
    object_list_sel = object_list[start_index:stop_index+1]

    for i, row in object_list_sel.iterrows():
        objectname = row['name']
        print(f"Starting fit {i}/{len(object_list)} for object: {objectname}")

        # read the setup
        filename = basedir + '/' + objectname + '/' + objectname + '_setup.yaml'
        try:
            infile = open(filename)
            setup = yaml.safe_load(infile)
            infile.close()
        except Exception as e:
            print("Could not read setupfile for {}".format(objectname))
            print(e)
            continue

        try:
            # get the photometry
            photbands, obs, obs_err = get_observations(setup)

            # do the SED fit
            results, samples, constraints, gridnames = fit_sed(setup, photbands, obs, obs_err)

            # add results to results dic
            all_results.setdefault('objectname', []).append(objectname)
            for k, v in results.items():
                all_results.setdefault(k, []).append(v[0])
                all_results.setdefault(k+'_err', []).append(( v[2] + v[3] ) / 2.0)

            # write the results
            write_results(setup, results, samples, obs, obs_err, photbands)

            # make any requested plots
            plot_results(setup, results, samples, constraints, gridnames, obs, obs_err, photbands)

        except Exception as e:
            print("Could not fit {}".format(objectname))
            print(e)

    all_results = pd.DataFrame(data=all_results)
    all_results.to_csv('results.csv')
    print(all_results)



def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('filename', action="store", type=str,
                        help='file containing a list with the names or coordinates of the systems to fit')
    parser.add_argument("-setup", dest='setup', type=str, default='single',
                        help="The path to the setup file to use, or a standard: 'single' or 'binary'. "
                             "(default = 'single'")
    parser.add_argument("-basedir", dest='basedir', type=str, default='./',
                        help="The directory where all photometry, setup and results will be stored. (default = ./)")
    parser.add_argument("--fit", dest='fit', action='store_true',
                        help="Fit the systems")
    parser.add_argument('--phot', dest='phot', action='store_true',
                        help='Obtain photometry of the systems')
    parser.add_argument('-start', dest='start_index', type=int, default=0,
                        help='Index of the object with which to start the fit, defaults to 0')
    parser.add_argument('-stop', dest='stop_index', type=int, default=None,
                        help='Index of the object with which to end the fitting, defaults to None (all objects are fitted).')
    args, variables = parser.parse_known_args()

    basedir = args.basedir
    default_setup = read_setup(args.setup)

    if not os.path.isdir(basedir):
        os.mkdir(basedir)

    object_data = pd.read_csv(args.filename)
    object_list = process_objects(object_data)

    # create the setup file for all systems
    prepare_setup(object_list, basedir, default_setup)

    # try to obtain photometry for all systems if requested
    if args.phot:
        prepare_photometry(object_list, basedir)

    # fit all systems if requested
    if args.fit:
        fit_seds(object_list, basedir, start_index=args.start_index, stop_index=args.stop_index)


from setuptools import setup, find_packages
import codecs
import os
import re


def read_meta():
    """
    Read the meta information from the nowcast.__init__.py file.  Assume UTF-8 encoding.
    """
    meta_path = os.path.join("speedyfit", "__init__.py")
    with codecs.open(os.path.join(os.path.abspath(os.path.dirname(__file__)), meta_path), "rb", "utf-8") as f:
        return f.read()


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta),
        META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


META_FILE = read_meta()

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
    "numpy",
    "emcee",
    "astropy>=5.1",
    "astroquery>=0.4.10",
    "pyvo>=1.4",
    "matplotlib",
    "pyaml",
    "corner",
    "tqdm",
    "scipy",
    "h5py",
    "pandas",
    "gaiadr3-zeropoint",
    "pytz",
]
setup(
    name="speedyfit",
    version=find_meta("version"),
    author=find_meta("author"),
    author_email=find_meta("email"),
    description=find_meta("description"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vosjo/speedyfit",
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    test_suite='pytest.collector',
    tests_require=['pytest'],
    entry_points={
        'console_scripts': ['speedyfit=speedyfit.main:main', 'speedyfit-batch=speedyfit.speedyfit_batch:main'],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Astronomy",
    ],
    python_requires='>=3.9',
)



import glob
import re
import os
import yaml
import numpy as np

from astropy.io import fits

from speedyfit import interpol, filters

__defaults__ = dict(grid='kurucz',
                    directory=os.environ.get('SPEEDYFIT_MODELS', None), )
defaults = __defaults__.copy()

# load a list of all available integrated grids
try:
    grid_description_file = os.path.join(defaults.get('directory', ''), 'grid_description.yaml')
    ifile = open(grid_description_file)
    grid_description = yaml.safe_load(ifile)
    ifile.close()
except:
    grid_description_file = os.path.join(defaults.get('directory', ''), 'grid_description.yaml')
    grid_description = {}


def check_grids(print_bands=False):
    """
    Check which atmospheric model grids are installed and can be used by speedyfit.

    :param print_bands: If True, also print the photometric pass bands that are available in the integrated grid.
    :type print_bands: bool
    """
    print("Checking which atmosphere models are available...")

    if defaults['directory'] is not None:
        print(f"Checking for models in {defaults['directory']}")
        print(f"Using grid description file: {grid_description_file}")
    else:
        print("SPEEDYFIT_MODELS environmental variable not set. CAN NOT find models!")
        print("Please point the SPEEDYFIT_MODELS variable to the directory where you stored models. On bash use:")
        print("\texport SPEEDYFIT_MODELS='<path to extracted atmosphere models>'")
        return

    if len(grid_description.keys()) == 0:
        print("grid_description.yaml file not found or no models included in the description file.")
        print("Please add a grid_description.yaml file in the SPEEDYFIT_MODELS directory.")
        print(f"SPEEDYFIT_MODELS directory is: {defaults.get('directory', '')}")
        files = glob.glob(os.path.join(defaults.get('directory', ''), '*'))
        print(f"Content of SPEEDYFIT_MODELS is: {files}")
        return

    for grid in grid_description.keys():
        print(grid)
        gridpath = get_grid_file(integrated=False, grid=grid)
        if os.path.isfile(gridpath):
            print("\t raw: available")
        else:
            print("\t raw: NOT FOUND")

        gridpath = get_grid_file(integrated=True, grid=grid)
        if os.path.isfile(gridpath):
            print("\t integrated: available")
        else:
            print("\t integrated: NOT FOUND")

        if print_bands:
            hdu = fits.open(gridpath)
            bands = set([b.split('.')[0] for b in hdu[1].data.dtype.names if '.' in b])
            for b in bands:
                print("\t - " + b)

        if 'info' in grid_description[grid]:
            print('\t info: ' + grid_description[grid]['info'])


def get_grid_file(integrated=False, **kwargs):
    grid = kwargs.get('grid', defaults['grid'])

    if os.path.isfile(grid):
        return grid

    if grid in grid_description:
        filename = grid_description[grid]['filename']
    else:
        raise ValueError(f'Grid name ({grid}) not recognized!')

    if integrated:
        filename = 'i' + filename + '_lawfitzpatrick2004_Rv3.10'

    directory = kwargs.get('directory', defaults['directory'])

    return os.path.join(directory, filename + '.fits')


def get_grid_ranges(**kwargs):
    grid = kwargs.get('grid', defaults['grid'])

    if grid == 'kurucz' or grid == 'kurucz2': # kurucz2 is necessary for backwards compatibility
        teff = (3500, 10000)
        logg = (4.32, 4.32)
        rad = (0.05, 2.5)

    elif grid == 'munari':
        teff = (3500, 20000)
        logg = (4.32, 4.32)
        rad = (0.05, 2.5)

    elif grid == 'tmap':
        teff = (20000, 80000)
        logg = (5.88, 5.88)
        rad = (0.01, 0.5)

    elif grid == 'koester':
        teff = (20000, 80000)
        logg = (5.88, 5.88)
        rad = (0.01, 0.5)

    elif grid == 'blackbody':
        teff = (20000, 80000)
        logg = (5.88, 5.88)
        rad = (0.01, 2.5)

    else:
        raise ValueError('Grid name ({}) not recognized!'.format(grid))

    return {'teff': teff, 'logg': logg, 'rad': rad}


def get_grid_dimensions(**kwargs):
    """
    Retrieve possible effective temperatures and gravities from a grid.

    E.g. kurucz, sdB, fastwind...

    :rtype: (ndarray,ndarray)
    :return: effective temperatures, gravities
    """
    gridfile = get_grid_file(**kwargs)
    ff = fits.open(gridfile)
    teffs = []
    loggs = []
    for hdu in ff[1:]:
        teffs.append(float(hdu.header['TEFF']))
        loggs.append(float(hdu.header['LOGG']))
    ff.close()

    # # -- maybe the fits extensions are not in right order...
    # matrix = np.vstack([np.array(teffs), np.array(loggs)]).T
    # matrix = numpy_ext.sort_order(matrix, order=[0, 1])
    # teffs, loggs = matrix.T

    return teffs, loggs


def load_grids(gridnames, pnames, limits, photbands):
    """
    prepares the integrated photometry grid by loading the grid and cutting it to the size
    given in limits.
    """
    grids = []
    for i, name in enumerate(gridnames):
        ind = '' if i == 0 else str(i + 1)

        axis_values, grid_pars, pixelgrid, grid_names = prepare_grid(photbands, name,
                                                                     teffrange=limits[pnames.index('teff' + ind)],
                                                                     loggrange=limits[pnames.index('logg' + ind)],
                                                                     ebvrange=limits[pnames.index('ebv')],
                                                                     variables=['teff', 'logg', 'ebv'])

        grids.append([axis_values, pixelgrid])

    return grids


def prepare_grid(photbands, gridname,
                 teffrange=(-np.inf, np.inf), loggrange=(-np.inf, np.inf),
                 ebvrange=(-np.inf, np.inf),
                 variables=['teff', 'logg', 'ebv'], **kwargs):
    flux = []
    grid_pars = []
    grid_names = np.array(variables)

    gridfilename = get_grid_file(integrated=True, grid=gridname)

    with fits.open(gridfilename) as ff:
        # -- make an alias for further reference
        ext = ff[1]

        # -- the grid is already cut here to limit memory usage
        keep = np.ones(len(ext.data), bool)
        for name in variables:
            # -- first find the closest actuall grid points
            low, high = locals()[name + 'range']
            lidx = np.abs(ext.data.field(name)[ext.data.field(name) <= low] - low).argmin()
            hidx = np.abs(ext.data.field(name)[ext.data.field(name) >= high] - high).argmin()
            low = ext.data.field(name)[ext.data.field(name) <= low][lidx]
            high = ext.data.field(name)[ext.data.field(name) >= high][hidx]

            # -- we need to be carefull for rounding errors
            in_range = (low <= ext.data.field(name)) & (ext.data.field(name) <= high)
            on_edge = np.allclose(ext.data.field(name), low) | np.allclose(ext.data.field(name), high)

            keep = keep & (in_range | on_edge)

        grid_pars = np.vstack([ext.data.field(name)[keep] for name in variables])
        flux = _get_flux_from_table(ext, photbands, include_Labs=True)[keep]

    flux = np.log10(flux)

    # -- create the pixeltype grid
    axis_values, pixelgrid = interpol.create_pixeltypegrid(grid_pars, flux.T)
    return axis_values, grid_pars.T, pixelgrid, grid_names


def get_itable_single(teff=None, logg=None, g=None, ebv=0.0, **kwargs):
    # -- check if logg or g is given
    if logg is None and not g is None:
        logg = np.log10(g)

    multiple = False
    if not hasattr(teff, '__iter__') or not hasattr(logg, '__iter__') or \
            not hasattr(ebv, '__iter__'):
        teff = np.array(teff)
        logg = np.array(logg)
        ebv = np.array(ebv)
        multiple = True

    # -- get the grid from the keyword, or prepare it if a grid name is given
    if isinstance(kwargs['grid'], str):
        axis_values, grid_pars, pixelgrid, grid_names = prepare_grid(kwargs['photbands'], kwargs['grid'],
                                                                     teffrange=(np.min(teff), np.max(teff)),
                                                                     loggrange=(np.min(logg), np.max(logg)),
                                                                     ebvrange=(np.min(ebv), np.max(ebv)),
                                                                     variables=['teff', 'logg', 'ebv'])
    else:
        axis_values, pixelgrid = kwargs['grid']

    p = np.vstack([teff, logg, ebv])

    values = interpol.interpolate(p, axis_values, pixelgrid)

    # -- switch logarithm to normal
    values = 10 ** values
    flux, Labs = values[:-1], values[-1]

    # -- Take radius into account when provided
    if 'rad' in kwargs:
        rad = np.array(kwargs['rad'])
        flux, Labs = flux * rad ** 2, Labs * rad ** 2

    if 'd' in kwargs:
        d = np.array(kwargs['d'])
        flux, Labs = flux / d ** 2, Labs * d ** 2

    if multiple:
        flux = flux.flatten()
        Labs = Labs.flatten()

    return flux, Labs


def get_itable(grid=[], **kwargs):
    values, parameters, components = {}, set(), set()
    for key in list(kwargs.keys()):
        if re.search("^(teff|logg|g|ebv|rad)\d?$", key):
            par, comp = re.findall("^(teff|logg|g|ebv|rad)(\d?)$", key)[0]
            values[key] = kwargs.pop(key)
            parameters.add(par)
            components.add(comp)

    # need to check here that grid is same length as components
    if isinstance(grid, str):
        grids = [grid]
    elif hasattr(grid, '__iter__') and isinstance(grid[0], str):
        grids = grid
    elif hasattr(grid, '__iter__') and hasattr(grid[0], '__iter__') and len(grid[0]) == 2:
        grids = grid
    else:
        grids = [grid]

    # -- If there is only one component, we can directly return the result
    if len(components) == 1:
        kwargs.update(values)
        fluxes, Labs = get_itable_single(grid=grids[0], **kwargs)
        return fluxes, {'L': Labs}

    fluxes, Labs = [], {}
    for i, (comp, grid) in enumerate(zip(components, grids)):
        kwargs_ = kwargs.copy()
        for par in parameters:
            kwargs_[par] = values[par + comp] if par + comp in values else values[par]

        f, L = get_itable_single(grid=grid, **kwargs_)

        fluxes.append(f)
        Labs['L' + comp] = np.sum(L, axis=0)

    fluxes = np.sum(fluxes, axis=0)
    return fluxes, Labs


def get_table_single(teff=None, logg=None, ebv=0.0, **kwargs):
    """
   No interpolating, just returns the closest gridpoint
   """

    # -- get the grid
    gridname = kwargs['grid']
    gridfilename = get_grid_file(integrated=False, grid=gridname)

    hdu = fits.open(gridfilename)

    teffs, loggs = np.zeros(len(hdu) - 1), np.zeros(len(hdu) - 1)
    for i in range(1, len(hdu)):
        teffs[i - 1] = hdu[i].header['TEFF']
        loggs[i - 1] = hdu[i].header['LOGG']

    dteff = abs(teffs - teff) / teff
    dlogg = abs(loggs - logg) / logg

    s = np.where(np.sqrt(dteff ** 2 + dlogg ** 2) == np.min(np.sqrt(dteff ** 2 + dlogg ** 2)))

    model = hdu[s[0][0] + 1].data
    wave, flux = model['wavelength'], model['flux']

    # -- Take radius into account when provided
    if 'rad' in kwargs:
        rad = np.array(kwargs['rad'])
        flux = flux * rad ** 2

    if 'd' in kwargs:
        d = np.array(kwargs['d'])
        flux = flux / d ** 2

    return wave, flux


def get_table(grid=[], **kwargs):
    """
   Returns the closest model atmosphere available in the grid. No interpolation is done!
   """
    values, parameters, components = {}, set(), set()
    for key in list(kwargs.keys()):
        if re.search("^(teff|logg|g|ebv|rad)\d?$", key):
            par, comp = re.findall("^(teff|logg|g|ebv|rad)(\d?)$", key)[0]
            values[key] = kwargs.pop(key)
            parameters.add(par)
            components.add(comp)

    # need to check here that grid is same length as components
    if hasattr(grid, '__iter__') and not isinstance(grid, str):
        grids = grid
    else:
        grids = [grid]

    # -- If there is only one component, we can directly return the result
    if len(components) == 1:
        kwargs.update(values)
        wave, flux = get_table_single(grid=grids[0], **kwargs)
        return wave, flux

    waves, fluxes = [], []
    for i, (comp, grid) in enumerate(zip(components, grids)):
        kwargs_ = kwargs.copy()
        for par in parameters:
            kwargs_[par] = values[par + comp] if par + comp in values else values[par]

        w, f = get_table_single(grid=grid, **kwargs_)

        waves.append(w)
        fluxes.append(f)

    # -- interpolate and combine the models
    wave = waves[0]
    flux = np.zeros_like(wave)
    for w, f in zip(waves, fluxes):
        flux += np.interp(wave, w, f)

    return wave, flux


def _get_flux_from_table(fits_ext, photbands, index=None, include_Labs=True):
    """
   Retrieve flux and flux ratios from an integrated SED table.

   @param fits_ext: fits extension containing integrated flux
   @type fits_ext: FITS extension
   @param photbands: list of photometric passbands
   @type photbands: list of str
   @param index: slice or index of rows to retrieve
   @type index: slice or integer
   @return: fluxes or flux ratios
   #@rtype: list
   """
    if index is None:
        index = slice(None)  # -- full range
    fluxes = []
    for photband in photbands:
        try:
            if not filters.is_color(photband):
                fluxes.append(fits_ext.data.field(photband)[index])
            else:
                system, color = photband.split('.')
                if '-' in color:
                    band0, band1 = color.split('-')
                    fluxes.append(fits_ext.data.field('%s.%s' % (system, band0))[index] /
                                  fits_ext.data.field('%s.%s' % (system, band1))[index])
                elif color == 'M1':
                    fv = fits_ext.data.field('STROMGREN.V')[index]
                    fy = fits_ext.data.field('STROMGREN.Y')[index]
                    fb = fits_ext.data.field('STROMGREN.B')[index]
                    fluxes.append(fv * fy / fb ** 2)
                elif color == 'C1':
                    fu = fits_ext.data.field('STROMGREN.U')[index]
                    fv = fits_ext.data.field('STROMGREN.V')[index]
                    fb = fits_ext.data.field('STROMGREN.B')[index]
                    fluxes.append(fu * fb / fv ** 2)
        except KeyError:
            print('Passband %s missing from table' % (photband))
            fluxes.append(np.nan * np.ones(len(fits_ext.data)))
    # -- possibly include absolute luminosity
    if include_Labs:
        fluxes.append(fits_ext.data.field("Labs")[index])
    fluxes = np.array(fluxes).T
    if index is not None:
        fluxes = fluxes
    return fluxes


def luminosity(wave, flux, radius=1.):
    """
    Calculate the bolometric luminosity of a model SED.

    Flux should be in cgs per unit wavelength (same unit as wave).
    The latter is integrated out, so it is of no importance. After integration,
    flux, should have units erg/s/cm2.

    Returned luminosity is in solar units.

    If you give radius=1 and want to correct afterwards, multiply the obtained
    Labs with radius**2.

    :param wave: model wavelengths
    :type wave: ndarray
    :param flux: model fluxes (Flam)
    :type flux: ndarray
    :param radius: stellar radius in solar units
    :type radius: float
    :return: total bolometric luminosity
    :rtype: float
    """
    Lsol_cgs = 3.846e33
    Rsol_cgs = 6.95508e10
    Lint = np.trapz(flux, x=wave)
    Labs = Lint * 4 * np.pi / Lsol_cgs * (radius * Rsol_cgs) ** 2
    return Labs



import os
import shutil
import logging
import configparser
import warnings

import numpy as np

from . import filters

from astroquery.simbad import Simbad
from astroquery.vizier import Vizier
from astroquery.utils.tap.core import TapPlus
import pyvo

from astropy import units as u
from astropy.coordinates.angles import Angle
from astropy.coordinates import SkyCoord
from astropy.io import ascii

from numpy.lib.recfunctions import append_fields

from zero_point import zpt #gaiadr3-zeropoint
zpt.load_tables()

FILE_DIR = os.path.dirname(os.path.abspath(__file__))
VIZIER_CAT = 'vizier_cats_phot.cfg'
TAP_CAT = 'tap_cats_phot.cfg'

# always request the distance of the observations to the search coordinates and sort on increasing distance
v = Vizier(columns=["*", '+_r'])

#-- read in catalog information
def load_photometry_catalogs():
    """
    Function to read the photometry catalogs: vizier_cats_phot.cfg and tap_cats_phot.cfg.
    The function will first look in the directory set with the SPEEDYFIT_MODELS environment variable. If no catalog
    files can be found in that directory, it will use the catalogs shipped with the speedyfit package.
    """
    user_dir = os.environ.get('SPEEDYFIT_MODELS', None)

    if user_dir is not None and os.path.exists(os.path.join(user_dir, VIZIER_CAT)):
        vizier_file = os.path.join(user_dir, VIZIER_CAT)
    else:
        vizier_file = os.path.join(FILE_DIR, VIZIER_CAT)

    if user_dir is not None and os.path.exists(os.path.join(user_dir, TAP_CAT)):
        tap_file = os.path.join(user_dir, TAP_CAT)
    else:
        tap_file = os.path.join(FILE_DIR, TAP_CAT)

    print(f"Going to read Vizier photometry catalog from {vizier_file}.")
    viz_info = configparser.ConfigParser()
    viz_info.optionxform = str # make sure the options are case sensitive
    with open(vizier_file) as ifile:
        viz_info.read_file(ifile)

    print(f"Going to read TAP photometry catalog from {tap_file}.")
    tap_info = configparser.ConfigParser()
    tap_info.optionxform = str  # make sure the options are case sensitive
    with open(tap_file) as ifile:
        tap_info.read_file(ifile)

    return viz_info, tap_info
VIZ_INFO, TAP_INFO = load_photometry_catalogs()

def copy_photometry_catalogs(overwrite=False):
    """
    Copy the two photometry catalogs: vizier_cats_phot.cfg and tap_cats_phot.cfg from the package source location
    to the SPEEDYFIT_MODELS directory, so that the user can edit them.
    """
    if os.environ.get('SPEEDYFIT_MODELS', None) is None:
        print("SPEEDYFIT_MODELS env variable is not set, can not copy the photometry catalogs.")
        return
    else:
        destination = os.environ.get('SPEEDYFIT_MODELS', None)

    vis_cat = os.path.join(FILE_DIR, VIZIER_CAT)
    if os.path.exists(vis_cat) and not overwrite:
        print(f"The file {vis_cat} already exists. To overwrite existing files use option --overwrite")
    elif (os.path.exists(vis_cat) and overwrite) or not os.path.exists(vis_cat):
        shutil.copy(vis_cat, destination)
        print(f"Copied {vis_cat} to {destination}")

    tap_cat = os.path.join(FILE_DIR, TAP_CAT)
    if os.path.exists(tap_cat) and not overwrite:
        print(f"The file {tap_cat} already exists. To overwrite existing files use option --overwrite")
    elif (os.path.exists(tap_cat) and overwrite) or not os.path.exists(tap_cat):
        shutil.copy(tap_cat, destination)
        print(f"Copied {tap_cat} to {destination}")


def get_coordinate(objectname):

    if objectname[0] == 'J' and objectname[1] != 'L':
        name = objectname[1:]
        if '+' in name:
            ra, dec = name.split('+')
            sign = '+'
        else:
            ra, dec = name.split('-')
            sign = '-'

        ra = "{}:{}:{}".format(ra[0:2], ra[2:4], ra[4:])
        dec = "{}{}:{}:{}".format(sign, dec[0:2], dec[2:4], dec[4:])

        ra = Angle(ra, unit='hour').degree
        dec = Angle(dec, unit='degree').degree

    else:
        data = Simbad.query_object(objectname)
        ra = Angle(data['ra'][0], unit='degree').degree
        dec = Angle(data['dec'][0], unit='degree').degree

    return ra, dec


def get_vizier_photometry(objectname, radius=5):

    catalogs = VIZ_INFO.sections()

    data = v.query_object(objectname, catalog=catalogs, radius=radius*u.arcsec)

    photometry = []

    for catalog in list(data.keys()):

        distance = (data[catalog]['_r'][0] * u.arcmin).to(u.arcsec).value

        if VIZ_INFO.has_option(catalog, 'bibcode'):
            bibcode = VIZ_INFO.get(catalog, 'bibcode')
        else:
            bibcode = '-'

        print(catalog)

        for band in VIZ_INFO.options(catalog):

            if '_unit' in band: continue
            if '_err' in band: continue
            if 'bibcode' in band: continue
            if '-' in band: continue

            bandname = VIZ_INFO.get(catalog, band)
            value = data[catalog][band][0]

            errkw = VIZ_INFO.get(catalog, band + '_err') if VIZ_INFO.has_option(catalog, band + '_err') else 'e_' + band
            try:
                err = data[catalog][errkw][0]
            except:
                err = 0.05

            if VIZ_INFO.has_option(catalog, band + '_unit'):
                unit = VIZ_INFO.get(catalog, band + '_unit')
            else:
                unit = data[catalog][band].unit

            photometry.append(( bandname, value, err, unit, distance, bibcode ))

    dtypes = [('band', '<U20'), ('meas', 'f8'), ('emeas', 'f8'), ('unit', '<U10'), ('distance', 'f8'), ('bibcode', '<U20')]
    photometry = np.array(photometry, dtype=dtypes)

    return photometry


def tap_query_vo(ra, dec, catalog):
    """
    info see:
    http://docs.g-vo.org/pyvo/html/page011.html
    """

    service = pyvo.dal.TAPService(catalog)

    table = TAP_INFO.get(catalog, 'table')
    print(table)

    keywords = ""
    bands = []
    for band in TAP_INFO.options(catalog):
        if 'table' in band: continue
        if 'rakw' in band: continue
        if 'deckw' in band: continue
        if '_unit' in band: continue
        if '_err' in band: continue
        if 'bibcode' in band: continue

        errkw = TAP_INFO.get(catalog, band + '_err') if TAP_INFO.has_option(catalog, band + '_err') else 'e_' + band

        keywords += band + ", " + errkw + ", "
        bands.append(band)

    if keywords[-2:] == ", ":
        keywords = keywords[0:-2]

    rakw = TAP_INFO.get(catalog, 'rakw') if TAP_INFO.has_option(catalog, 'rakw') else 'raj2000'
    deckw = TAP_INFO.get(catalog, 'deckw') if TAP_INFO.has_option(catalog, 'deckw') else 'dej2000'

    # -- first try to query with distance
    query = """SELECT 
         DISTANCE(POINT('ICRS', {rakw:}, {deckw}),
                  POINT('ICRS', {ra:}, {dec:})) AS dist, {kws:}
         FROM {table:} AS m
         WHERE 
            1=CONTAINS(POINT('ICRS', {rakw:}, {deckw}),
                     CIRCLE('ICRS', {ra:}, {dec:}, 0.005 ))
         ORDER BY dist""".format(ra=ra, dec=dec, table=table, rakw=rakw, deckw=deckw, kws=keywords)

    results = service.run_sync(query).to_table()

    if len(results) == 0:
        dtypes = [('band', '<U20'), ('meas', 'f8'), ('emeas', 'f8'), ('unit', '<U10'), ('distance', 'f8'), ('bibcode', '<U20')]
        return np.array([], dtype=dtypes)

    # -- get the distance from the query result or calculate it.
    if 'dist' in results.dtype.names:
        distance = (results['dist'][0] * u.degree).to(u.arcsec).value
    else:
        c1 = SkyCoord(ra * u.degree, dec * u.degree)
        distance = SkyCoord(results['ra'][0] * u.degree, results['de'][0] * u.degree).separation(c1).to(u.arcsec).value

    bibcode = TAP_INFO.get(catalog, 'bibcode')

    # -- get the photometric measurements, errors and units
    photometry = []
    for band in bands:
        bandname = TAP_INFO.get(catalog, band)
        value = results[band][0]

        errkw = TAP_INFO.get(catalog, band + '_err') if TAP_INFO.has_option(catalog, band + '_err') else 'e_' + band
        err = results[errkw][0]

        if TAP_INFO.has_option(catalog, band + '_unit'):
            unit = TAP_INFO.get(catalog, band + '_unit')
        else:
            unit = results[band].unit

        photometry.append((bandname, value, err, unit, distance, bibcode))

    dtypes = [('band', '<U20'), ('meas', 'f8'), ('emeas', 'f8'), ('unit', '<U10'), ('distance', 'f8'), ('bibcode', '<U20')]
    photometry = np.array(photometry, dtype=dtypes)

    return photometry


def tap_query(ra, dec, catalog):

    tp = TapPlus(url=catalog)

    table = TAP_INFO.get(catalog, 'table')
    print(table)

    keywords = ""
    bands = []
    for band in TAP_INFO.options(catalog):
        if 'table' in band: continue
        if 'rakw' in band: continue
        if 'deckw' in band: continue
        if '_unit' in band: continue
        if '_err' in band: continue
        if 'bibcode' in band: continue

        errkw = TAP_INFO.get(catalog, band + '_err') if TAP_INFO.has_option(catalog, band + '_err') else 'e_' + band

        keywords += band + ", " + errkw + ", "
        bands.append(band)

    if keywords[-2:] == ", ":
        keywords = keywords[0:-2]

    rakw = TAP_INFO.get(catalog, 'rakw') if TAP_INFO.has_option(catalog, 'rakw') else 'raj2000'
    deckw = TAP_INFO.get(catalog, 'deckw') if TAP_INFO.has_option(catalog, 'deckw') else 'dej2000'

    try:
        #-- first try to query with distance
        query = """SELECT 
      DISTANCE(POINT('ICRS', {rakw:}, {deckw}),
               POINT('ICRS', {ra:}, {dec:})) AS dist, {kws:}
      FROM {table:} AS m
      WHERE 
         1=CONTAINS(POINT('ICRS', {rakw:}, {deckw}),
                  CIRCLE('ICRS', {ra:}, {dec:}, 0.005 ))
      ORDER BY dist""".format(ra=ra, dec=dec, table=table, rakw=rakw, deckw=deckw, kws=keywords)

        job = tp.launch_job(query)

        results = job.get_results()

    except Exception as e:

        print("problem", e)

        #-- not all catalogs accept a distance sorted query, we have to hope that the correct star is returned.
        query = """SELECT {rakw:} as RA, {deckw} as DE, {kws:}
      FROM {table:} AS m
      WHERE 
         1=CONTAINS(POINT('ICRS', {rakw:}, {deckw}),
                  CIRCLE('ICRS', {ra:}, {dec:}, 0.005 ))
      """.format(ra=ra, dec=dec, table=table, rakw=rakw, deckw=deckw, kws=keywords)

        job = tp.launch_job(query)

        results = job.get_results()

    if len(results) == 0:
        dtypes = [('band', 'a20'), ('meas', 'f8'), ('emeas', 'f8'), ('unit', 'a10'), ('distance', 'f8'), ('bibcode', 'a20')]
        return np.array([], dtype=dtypes)

    #-- get the distance from the query result or calculate it.
    if 'dist' in results.dtype.names:
        distance = (results['dist'][0] * u.degree).to(u.arcsec).value
    else:
        c1 = SkyCoord(ra * u.degree, dec * u.degree)
        distance = SkyCoord(results['ra'][0] * u.degree, results['de'][0] * u.degree).separation(c1).to(u.arcsec).value


    bibcode = TAP_INFO.get(catalog, 'bibcode')

    #-- get the photometric measurements, errors and units
    photometry = []
    for band in bands:
        bandname = TAP_INFO.get(catalog, band)
        value = results[band][0]

        errkw = TAP_INFO.get(catalog, band + '_err') if TAP_INFO.has_option(catalog, band + '_err') else 'e_' + band
        err = results[errkw][0]

        if TAP_INFO.has_option(catalog, band + '_unit'):
            unit = TAP_INFO.get(catalog, band + '_unit')
        else:
            unit = results[band].unit

        photometry.append(( bandname, value, err, unit, distance, bibcode))

    dtypes = [('band', '<U20'), ('meas', 'f8'), ('emeas', 'f8'), ('unit', '<U10'), ('distance', 'f8'), ('bibcode', '<U20')]
    photometry = np.array(photometry, dtype=dtypes)

    return photometry


def get_tap_photometry(ra, dec):
    catalogs = TAP_INFO.sections()

    dtypes = [('band', '<U20'), ('meas', 'f8'), ('emeas', 'f8'), ('unit', '<U10'), ('distance', 'f8'), ('bibcode', '<U20')]
    photometry = np.array([], dtype=dtypes)

    for catalog in catalogs:

        phot_ = tap_query_vo(ra, dec, catalog)
        photometry = np.hstack([photometry, phot_])

    return photometry


def get_photometry(objectname, filename=None):

    #-- get the object coodinates
    ra, dec = get_coordinate(objectname)

    #-- query tap catalogs
    photometry = get_tap_photometry(ra, dec)

    #-- query Vizier catalogs
    photometry_ = get_vizier_photometry("{} {}".format(ra, dec))

    print(photometry)
    print(photometry_)

    photometry = np.hstack([photometry, photometry_])

    #-- convert magnitudes to fluxes
    wave, flux, err = [], [], []
    for band, meas, emeas, unit in zip(photometry['band'], photometry['meas'], photometry['emeas'], photometry['unit']):
        if np.isnan(emeas) or emeas < 0: emeas = 0.02

        f_, e_ = filters.mag2flux(meas, emeas, band)

        wave.append(filters.eff_wave(band))
        flux.append(f_)
        err.append(e_)

    photometry = append_fields(photometry, ['flux', 'eflux'], data=[flux, err],
                               dtypes=['f8', 'f8'], usemask=False)

    if filename is not None:
        ascii.write(photometry, filename, format='fixed_width', overwrite=True)

    return photometry


def get_parallax(objectname, ra=None, dec=None, radius=5):

    v_gaia = Vizier(columns=["Plx", "e_Plx", '+_r', 'Gmag', 'nueff', 'pscol', 'ELAT', 'Solved'])

    # try the query with the objectname first
    data = v_gaia.query_object(objectname, catalog=['I/350/gaiaedr3'], radius=radius*u.arcsec)

    # if no data is returned retry the query with the coordinates if possible
    if len(data) == 0 and ra is not None and dec is not None:

        return None, None

    data = data['I/350/gaiaedr3'][0]

    plx, plx_e = data['Plx'], data['e_Plx']

    if not data['pscol']:
        data['pscol'] = 0

    # apply parallax zero point correction of Lindgren. If not possible, use the average offset.
    # https://arxiv.org/pdf/2012.01742.pdf
    # https://gitlab.com/icc-ub/public/gaiadr3_zeropoint
    try:
        zp = zpt.get_zpt(data['Gmag'], data['nueff'], data['pscol'], data['ELAT'], data['Solved'])
    except Exception:
        warnings.warn("Could not calculate the parallax zero point offset based on Lindgren+2020, using average")
        zp = 0.02
    plx -= zp

    return plx, plx_e

